/*
	Copyright (C) 2005  Brendan G Bohannon (aka: cr88192)
	
	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.
	
	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
	MA 02111-1307, USA
	
	Email: cr88192@hotmail.com
*/

/*
	More complicated encoder/decoder centered around files.
*/

#include <lz2a.h>


unsigned int LZ2A_FileAdler32(FILE *fd, ulsz *rsz)
{
	byte *buf;
	unsigned int crc;
	ulsz sz;
	int i;

	buf=malloc(65536);

	sz=0;
	crc=1;
	while(!feof(fd))
	{
		i=fread(buf, 1, 65536, fd);
		if(i<=0)break;

		crc=PDRLELZ_DataAdler32(buf, i, crc);
		sz+=i;
	}
	free(buf);

	if(rsz)*rsz=sz;

	return(crc);
}

int LZ2A_WriteHeaderFD(FILE *infd, FILE *outfd, ulsz *rsz)
{
	byte *cs, *ct, *ce;
	byte *buf;
	unsigned int crc;
	ulsz sz;
	int i, j;

	crc=LZ2A_FileAdler32(infd, &sz);

	if(sz<lz2a_dictsz)
	{
		i=log2up(sz);
		lz2a_dictsz=1<<i;

		if(lz2a_verbose)fprintf(stderr, "Used dictsz %d\n", i);
	}

	buf=malloc(32768);
	ct=buf;

#ifdef WHUFF_NOLONGLONG
	//low order size
	*ct++=0x01;
	*ct++=6;
	*ct++=sz&0xFF;
	*ct++=(sz>>8)&0xFF;
	*ct++=(sz>>16)&0xFF;
	*ct++=(sz>>24)&0xFF;
#else
	//high order size
	*ct++=0x01;

	if(sz>(1<<30))
	{
		*ct++=10;
		*ct++=sz&0xFF;
		*ct++=(sz>>8)&0xFF;
		*ct++=(sz>>16)&0xFF;
		*ct++=(sz>>24)&0xFF;
		*ct++=(sz>>32)&0xFF;
		*ct++=(sz>>40)&0xFF;
		*ct++=(sz>>48)&0xFF;
		*ct++=(sz>>56)&0xFF;
	}else
	{
		*ct++=6;
		*ct++=sz&0xFF;
		*ct++=(sz>>8)&0xFF;
		*ct++=(sz>>16)&0xFF;
		*ct++=(sz>>24)&0xFF;
	}
#endif

	//checksum
	*ct++=0x02;
	*ct++=6;
	*ct++=crc&0xFF;
	*ct++=(crc>>8)&0xFF;
	*ct++=(crc>>16)&0xFF;
	*ct++=(crc>>24)&0xFF;

	//dictionary size
	*ct++=0x03;
	*ct++=6;
	*ct++=log2up(lz2a_dictsz);
	*ct++=lz2a_runbits;
	*ct++=lz2a_distbits;
	*ct++=warith2_ctxbits;

	//other bits of info can be put here


	//algo magic
	fputc('L', outfd);
	fputc('Z', outfd);
	fputc('2', outfd);
	fputc('A', outfd);

	fputc(LZ2A_VERSION, outfd);	//algo version a.b, eg: 0.1
	fputc(0x00, outfd);		//reserved, set to 0

	//rest of header info size
	i=ct-buf;
	j=i+8;
	fputc(j&0xFF, outfd);
	fputc((j>>8)&0xFF, outfd);

	fwrite(buf, 1, i, outfd);
	free(buf);

	//marker for begining of encoded stream
	fputc('S', outfd);
	fputc('T', outfd);
	fputc('R', outfd);
	fputc('M', outfd);


	if(rsz)*rsz=sz;
	return(0);
}

int LZ2A_CompLZ2A_Setup(byte *in, byte *out, int sz,
	int *hash, int *hcbuf, int *num)
{
	int cnt[256], rnum[256];
	byte v[16];
	int i;

	PDRLELZ_StatBuffer(cnt, num, in, sz);
	PDRLELZ_SortStats(cnt, num);
	for(i=0; i<256; i++)rnum[num[i]]=i;
	for(i=0; i<16; i++)v[i]=num[0xF0+i];

	WHUFF_SetupEncode(out);
	LZ2A_SetupContexts();

	for(i=0; i<16; i++)
		WArith2_EncodeSymbol(lz2a_teststr[i]);

	//version dependent header
	//16 reserved codes
	for(i=0; i<16; i++)
		WArith2_EncodeSymbol(v[i]);

}

int LZ2A_CompLZ2A_Core(byte *in, byte *cs, byte *ce, byte *out,
	int sz, int *hash, int *hcbuf, int *num)
{
	int rnum[256];

	int i, j, k, l, t;
	int v[16];
	int li, lj, lk, ll, nl;
	int bl, bn, bi, bnl;	//match run/offset
	int lbl, lbn;	//last match run/offset
	int rl, rv, rc;	//run len/value

	for(i=0; i<256; i++)rnum[num[i]]=i;

	if(lz2a_verbose)fprintf(stderr, "encode\n");

	t=1024;
	rc=0;

	while(1)
	{
		if(cs>=ce)break;

#if 1
		if(!t && lz2a_verbose)
		{
			i=cs-in;
			j=ce-in;
			k=huff_winbuf-out;

			fprintf(stderr, "%d/%d %2.2f%% %2.2f%%\r",
				i, j, (100.0*i)/j, (100.0*k)/i);
			fflush(stderr);
			t=1024;
		}else t--;
#endif

#if 1
		//try looking up a string
		if((cs+5)<ce)
		{
			bl=0;
			bn=0;
			bi=0;
			bnl=0;

			LZ2A_LookupString(in, cs, ce, hash, hcbuf,
				&bl, &bn, &bi, lz2a_dictsz, RLELZ_RUNSZ-1);

			while(bi--)
			{
				LZ2A_UpdateHashChar(in, cs, ce, hash, hcbuf);
				if(rnum[*cs]>=0xF0)
					WArith2_EncodeSymbol(num[255]);
				WArith2_EncodeSymbol(*cs++);
			}

#if 1
			if(bl>5)
//			if(bl)
			{
				lj=((cs-in)-bn);
//				if(!rc)if(lz2a_verbose)fprintf(stderr, "%d %d %d\n",
//					cs-in, bl, lj);
				rc++;

				WArith2_EncodeSymbol(num[254]);
				LZ2A_EncodeReference(bl, lj);

				LZ2A_UpdateHash(in, cs, ce, hash, hcbuf,
					bl, bn);
				cs+=bl;
				continue;
			}
#endif
		}
#endif
		LZ2A_UpdateHashChar(in, cs, ce, hash, hcbuf);

		if(rnum[*cs]>=0xF0)
			WArith2_EncodeSymbol(num[255]);
		WArith2_EncodeSymbol(*cs++);
	}

	if(lz2a_verbose)fprintf(stderr, "\n");

	return(0);
}

int LZ2A_EncodeFile(FILE *infd, FILE *outfd)
{
	int num[256];

	byte *ibuf, *obuf;
	int *hash, *hcbuf;
	ulsz szi, szo, sz;
	int i, j, k;

	lz2a_distbits=log2up(lz2a_dictsz);

	szi=0;
	szo=0;

	if(lz2a_verbose)fprintf(stderr, "header\n");
	LZ2A_WriteHeaderFD(infd, outfd, &sz);

	if(lz2a_verbose)fprintf(stderr, "alloc\n");
	ibuf=malloc(lz2a_dictsz*2);
	obuf=malloc(lz2a_dictsz*2);
	memset(ibuf, 0, lz2a_dictsz*2);
	memset(obuf, 0, lz2a_dictsz*2);

	hash=malloc(RLELZ_HASHSZ*sizeof(int));
	hcbuf=malloc(lz2a_dictsz*2*sizeof(int));
	memset(hash, 0, RLELZ_HASHSZ*sizeof(int));
	memset(hcbuf, 0, lz2a_dictsz*2*sizeof(int));

	if(lz2a_verbose)fprintf(stderr, "setup\n");

	fseek(infd, 0, 0);
	i=fread(ibuf, 1, lz2a_dictsz, infd);
	szi+=i;
	LZ2A_CompLZ2A_Setup(ibuf, obuf, lz2a_dictsz, hash, hcbuf, num);

	if(lz2a_verbose)fprintf(stderr, "primer %d\n", i);

	//first window, primer
	LZ2A_CompLZ2A_Core(ibuf, ibuf, ibuf+i, obuf,
		i, hash, hcbuf, num);
	j=huff_winbuf-obuf;
	fwrite(obuf, 1, j, outfd);
	szo+=j;
	huff_winbuf=obuf;

	k=1;
	if(lz2a_verbose)fprintf(stderr, "block %d %d/%d\n",
		i, k, (sz+(lz2a_dictsz-1))/lz2a_dictsz);
	k++;

	while(1)
	{
		i=fread(ibuf+lz2a_dictsz, 1, lz2a_dictsz, infd);
		if(i<=0)break;
		szi+=i;

		LZ2A_CompLZ2A_Core(ibuf, ibuf+lz2a_dictsz,
			ibuf+lz2a_dictsz+i, obuf,
			i, hash, hcbuf, num);
		j=huff_winbuf-obuf;
		fwrite(obuf, 1, j, outfd);
		szo+=j;
		huff_winbuf=obuf;

		if(lz2a_verbose)fprintf(stderr, "block %d %d/%d\n",
			i, k, (sz+(lz2a_dictsz-1))/lz2a_dictsz);
		k++;

		//shift over whole mass
		memcpy(ibuf, ibuf+lz2a_dictsz, lz2a_dictsz);

		for(i=0; i<lz2a_dictsz; i++)
		{
			j=hcbuf[i+lz2a_dictsz]-lz2a_dictsz;
			if(j<0)j=0;
			hcbuf[i]=j;
		}
		for(i=0; i<RLELZ_HASHSZ; i++)
		{
			j=hash[i]-lz2a_dictsz;
			if(j<0)j=0;
			hash[i]=j;
		}
	}

	//finish up
	WArith2_EncodeSymbol(num[253]);

	WArith2_EncodeSymbol('E');
	WArith2_EncodeSymbol('O');
	WArith2_EncodeSymbol('M');
	WArith2_EncodeSymbol(' ');

	WHUFF_FlushWBits();
	j=huff_winbuf-obuf;
	fwrite(obuf, 1, j, outfd);
	szo+=j;

	if(lz2a_verbose)fprintf(stderr, "\n");
	if(lz2a_verbose)fprintf(stderr, "%d %d %2.2f%%\n",
		(long)szi, (long)szo, ((100.0*szo)/szi));

	return(0);
}


int LZ2A_ReadHeaderFD(FILE *fd,
	ulsz *rsz, unsigned int *rcrc)
{
	int i, j, k;
	unsigned int crc;
	ulsz sz;

	i=fgetc(fd);
	i+=fgetc(fd)<<8;
	i+=fgetc(fd)<<16;
	i+=fgetc(fd)<<24;
	if(i!=('L'+('Z'<<8)+('2'<<16)+('A'<<24)))
	{
		fprintf(stderr, "Invalid algo magic\n");
		return(-1);
	}

	i=fgetc(fd);
	if(i!=LZ2A_VERSION)
	{
		fprintf(stderr, "Wrong LZ2A version\n");
		return(-1);
	}

	i=fgetc(fd);
	if(i)
	{
		fprintf(stderr, "Reserved value!=0\n");
		return(-1);
	}

	k=fgetc(fd);
	k+=fgetc(fd)<<8;
	if((k<8) || (k>32767))
	{
		fprintf(stderr, "Invalid Header Size %d\n", k);
		return(-1);
	}

	while(ftell(fd)<k)
	{
		i=fgetc(fd);
		if(i==0x01)
		{
			sz=0;
			j=0;
			i=fgetc(fd)-2;
			while((i--)>0)
			{
				sz+=fgetc(fd)<<j;
				j+=8;
			}

			if(lz2a_verbose)fprintf(stderr, "Size %d\n", (int)sz);
			if(rsz)*rsz=sz;
			continue;
		}
		if(i==0x02)
		{
			crc=0;
			j=0;
			i=fgetc(fd)-2;
			while((i--)>0)
			{
				crc+=fgetc(fd)<<j;
				j+=8;
			}

			if(lz2a_verbose)fprintf(stderr, "CRC %08X\n", crc);

			if(rcrc)*rcrc=crc;
			continue;
		}

		if(i==0x03)
		{
			i=fgetc(fd)-3;

			j=fgetc(fd);
			if(lz2a_verbose)fprintf(stderr, "Window %d (%d bit)\n", 1<<j, j);
			lz2a_dictsz=1<<j;

			if(i)
			{
				j=fgetc(fd);
				if(lz2a_verbose)fprintf(stderr, "Run Bits %d\n", j);
				if(j)lz2a_runbits=j;
				i--;
			}

			if(i)
			{
				j=fgetc(fd);
				if(lz2a_verbose)fprintf(stderr, "Dist Bits %d\n", j);
				if(j)lz2a_distbits=j;
				i--;
			}

			if(i)
			{
				j=fgetc(fd);
				if(lz2a_verbose)fprintf(stderr, "Context Bits %d\n", j);
				if(j)warith2_ctxbits=j;
				if(j)warith2_ctxmask=(1<<warith2_ctxbits)-1;
				i--;
			}

			while((i--)>0)fgetc(fd);

			continue;
		}

		i=fgetc(fd);
		while((i--)>0)fgetc(fd);
	}

	fseek(fd, k, 0);


	i=fgetc(fd);
	i+=fgetc(fd)<<8;
	i+=fgetc(fd)<<16;
	i+=fgetc(fd)<<24;
	if(i!=('S'+('T'<<8)+('R'<<16)+('M'<<24)))
	{
		if(lz2a_verbose)fprintf(stderr, "Invalid stream magic\n");
		return(-1);
	}

	return(ftell(fd));
}

int LZ2A_DecompLZ2A_Core(byte *out, int osz, int *num)
{
	int i, j, k, l, ll;
	byte *ct, *s2;

	if(!osz)return(0);

	ct=out;
	while((ct-out)<osz)
	{
		j=WArith2_DecodeSymbol();
		i=num[j];
		if(i)
		{
			if(i==0xFF)
			{
				j=WArith2_DecodeSymbol();
				*ct++=j;
				continue;
			}
			if(i==253)
			{
				break;
			}

			if(i==0xFE)
			{
				LZ2A_DecodeReference(&l, &k);

				ll=l;
				while(k<l)
				{
					memcpy(ct, ct-k, k);
					ct+=k;
					l-=k;
				}

				memcpy(ct, ct-k, l);
				ct+=l;
				continue;
			}

			if(lz2a_verbose)fprintf(stderr, "Invalid control code\n");
//			return(-1);
		}

		*ct++=j;
		continue;
	}

	return(ct-out);
}

int LZ2A_DecodeFile(FILE *infd, FILE *outfd)
{
	int num[256];

	byte *ibuf, *obuf;
	ulsz sz;
	unsigned int crc, crc2;
	ulsz szi, szo;
	int i, j, k, l;

	szi=0;
	szo=0;

	LZ2A_ReadHeaderFD(infd, &sz, &crc);

	ibuf=malloc(lz2a_dictsz*2);
	obuf=malloc((lz2a_dictsz*2)+(2*RLELZ_RUNSZ));
	memset(ibuf, 0, lz2a_dictsz*2);
	memset(obuf, 0, lz2a_dictsz*2);

	i=fread(ibuf, 1, 2*lz2a_dictsz, infd);
	szi+=i;
	if(lz2a_verbose)fprintf(stderr, "consumed %d bytes\n", i);

	WHUFF_SetupDecode(ibuf);
	LZ2A_SetupContexts();

	for(i=0; i<16; i++)
	{
		j=WArith2_DecodeSymbol();

		if(j!=lz2a_teststr[i])
		{
			if(lz2a_verbose)fprintf(stderr, "Invalid test string"
				"(possible encoding/decoding issue)\n", j);
			return(-1);
		}
	}

	for(i=0; i<256; i++)
		num[i]=0;
	for(i=0; i<16; i++)
	{
		j=WArith2_DecodeSymbol();
		num[j]=i+0xF0;
	}

	//primer
	i=LZ2A_DecompLZ2A_Core(obuf, lz2a_dictsz, num);
//	crc2=PDRLELZ_DataAdler32(obuf, i, 1);
	crc2=1;
	l=1;

	if(lz2a_verbose)fprintf(stderr, "block %d/%d\n", l,
		(sz+(lz2a_dictsz-1))/lz2a_dictsz);
	l++;

	if(i>=lz2a_dictsz)
	{
		while(1)
		{
			//shift input
			j=huff_winbuf-ibuf;
			memcpy(ibuf, ibuf+j, (2*lz2a_dictsz)-j);
			k=fread(ibuf+((2*lz2a_dictsz)-j), 1, j, infd);
			huff_winbuf=ibuf;
			szi+=j;
			if(lz2a_verbose)fprintf(stderr, "consumed %d bytes, shifted %d\n",
				k, j);

			//decode more
			k=LZ2A_DecompLZ2A_Core(obuf+i,
				(2*lz2a_dictsz)-i, num);

			//output
			fwrite(obuf, 1, i, outfd);
			crc2=PDRLELZ_DataAdler32(obuf, i, crc2);
			szo+=i;
			if(lz2a_verbose)fprintf(stderr, "excreeted %d bytes\n", i);

			//shift output
			memcpy(obuf, obuf+i, k);
			i=k;


			if(lz2a_verbose)fprintf(stderr, "block %d/%d\n", l,
				(sz+(lz2a_dictsz-1))/lz2a_dictsz);
			l++;

			if(i<lz2a_dictsz)break;
		}
	}

	fwrite(obuf, 1, i, outfd);
	crc2=PDRLELZ_DataAdler32(obuf, i, crc2);
	szo+=i;
	if(lz2a_verbose)fprintf(stderr, "excreeted %d bytes\n", i);

#if 1
	j=WArith2_DecodeSymbol();
	if(num[j]==253)j=WArith2_DecodeSymbol();
	j+=WArith2_DecodeSymbol()<<8;
	j+=WArith2_DecodeSymbol()<<16;
	j+=WArith2_DecodeSymbol()<<24;
	if(j!=('E'+('O'<<8)+('M'<<16)+(' '<<24)))
	{
		if(lz2a_verbose)fprintf(stderr, "Missing 'EOM ' tag %08X"
			"(possible encoding/decoding issue)\n", j);
	}
#endif

	if(lz2a_verbose)fprintf(stderr, "\n");
	if(lz2a_verbose)fprintf(stderr, "%d %d %2.2f%%\n",
		(long)szi, (long)szo, ((100.0*szo)/szi));

	if(szo!=sz)if(lz2a_verbose)fprintf(stderr, "size mismatch %d!=%d\n",
		(int)sz, (int)szo);
	if(crc!=crc2)if(lz2a_verbose)fprintf(stderr, "crc mismatch %08X!=%08X\n",
		crc, crc2);

	free(ibuf);
	free(obuf);
}
